@@ -1,5 +1,5 @@
 /***************************************************************************
- *   Copyright (C) 2010 by Hans-Peter Deifel                               *
+ *   Copyright (C) 2012 by Hans-Peter Deifel                               *
  *   hpdeifel@gmx.de                                                       *
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
@@ -18,6 +18,8 @@
  *   Boston, MA 02110-1301 USA.                                            *
  ***************************************************************************/
 
+#include "config.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -28,20 +30,22 @@
 #include <unistd.h>
 #include <math.h>
 #include <sys/ioctl.h>
+#include <fnmatch.h>
+#include <errno.h>
+#include <dirent.h>
+#include <sys/stat.h>
 
-#include <PDFDoc.h>
-#include <goo/GooString.h>
-#include <TextOutputDev.h>
-#include <GlobalParams.h>
+#include <cpp/poppler-document.h>
+#include <cpp/poppler-page.h>
 
-#include "config.h"
+#ifdef HAVE_UNAC
+#include <unac.h>
+#endif
 
-#define BUFFER_SIZE 1024
+#include <memory>
 
-static char *filename_color;
-static char *pagenum_color;
-static char *highlight_color;
-static char *seperator_color;
+#include "output.h"
+#include "exclude.h"
 
 /* set this to 1 if any match was found. Used for the exit status */
 int found_something = 0;
@@ -49,19 +53,34 @@ int found_something = 0;
 /* default options */
 
 int ignore_case = 0;
-int color = 1;
+int f_recursive_search = 0;
 /* characters of context to put around each match.
  * -1 means: print whole line
  * -2 means: try to fit context into terminal */
 int context = -2;
 int line_width = 80;
-int print_filename = -1;
-int print_pagenum = 0;
 int count = 0;
 int quiet = 0;
 
-#define HELP_OPTION 1
-#define COLOR_OPTION 2
+#ifdef HAVE_UNAC
+int use_unac = 0;
+#endif
+
+struct outconf outconf = {
+	-1,			/* filename */
+	0,			/* pagenum */
+	1,			/* color */
+};
+
+enum {
+	HELP_OPTION,
+	COLOR_OPTION,
+	EXCLUDE_OPTION,
+	INCLUDE_OPTION,
+#ifdef HAVE_UNAC
+	UNAC_OPTION,
+#endif
+};
 
 struct option long_options[] =
 {
@@ -72,241 +91,99 @@ struct option long_options[] =
 	{"count", 0, 0, 'c'},
 	{"color", 1, 0, COLOR_OPTION},
 	{"context", 1, 0, 'C'},
+	{"recursive", 0, 0, 'r'},
+	{"recursive", 0, 0, 'R'},
+	{"exclude", 1, 0, EXCLUDE_OPTION},
+	{"include", 1, 0, INCLUDE_OPTION},
 	{"help", 0, 0, HELP_OPTION},
 	{"version", 0, 0, 'V'},
 	{"quiet", 0, 0, 'q'},
+#ifdef HAVE_UNAC
+	{"unac", 0, 0, UNAC_OPTION},
+#endif
 	{0, 0, 0, 0}
 };
 
-struct stream {
-	int bufsize;
-	char *buf;
-	int charpos;
-};
+ExcludeList excludes;
+ExcludeList includes;
 
-struct stream* reset_stream(struct stream *s)
+void regmatch_to_match(const regmatch_t match, int index, struct match *mt)
 {
-	s->bufsize = BUFFER_SIZE;
-	free(s->buf);
-	s->buf = (char*) malloc(BUFFER_SIZE);
-	s->charpos = 0;
-
-	return s;
+	mt->start = match.rm_so + index;
+	mt->end   = match.rm_eo + index;
 }
 
-struct stream* make_stream()
+#ifdef HAVE_UNAC
+/* convenience layer over libunac. The result has to be freed with
+ * simple_unac_free */
+char *simple_unac(char *string)
 {
-	struct stream *s = (struct stream*) malloc(sizeof(struct stream));
+	if (!use_unac)
+		return string;
 
-	s->buf = NULL;
+	char *res = NULL;
+	size_t reslen = 0;
 
-	return reset_stream(s);
-}
-
-void maybe_update_buffer(struct stream *s, int len)
-{
-	if (s->charpos + len >= s->bufsize) {
-		/* resize buffer */
-		do
-			s->bufsize = s->bufsize * 2;
-		while (s->charpos + len >= s->bufsize);
-
-		s->buf = (char*) realloc(s->buf, s->bufsize);
-	}
-}
-
-void write_to_stream(void *s, char *text, int length)
-{
-	struct stream *stream = (struct stream*) s;
-
-	maybe_update_buffer(stream, length);
-	strncpy(stream->buf+stream->charpos,
-			text, length);
-	stream->charpos += length;
-}
-
-void start_color(const char *colorcode)
-{
-	if (color)
-		printf("\33[%sm\33[K", colorcode);
-}
-
-void end_color()
-{
-	if (color)
-		printf("\33[m\33[K");
-}
-
-#define with_color(color, code) \
-	do { \
-		start_color(color); \
-		{ \
-			code; \
-		} \
-		end_color(); \
-	} while (0);
-
-void putsn(char *string, int from, int to)
-{
-	for (; from < to; from++)
-		putchar(string[from]);
-}
-
-int next_word_left(char *string, int index)
-{
-	int i = 0;
-	int in_whitespace;
-
-	if (index < 0 || string[index] == '\n')
-		return -1;
-
-	in_whitespace = isspace(string[index]);
-	while (index >= 0 && string[index] != '\n') {
-		if (in_whitespace) {
-			if (!isspace(string[index]))
-				in_whitespace = 0;
-		} else {
-			if (isspace(string[index]))
-				break;
-		}
-		i++;
-		index--;
+	if (unac_string("UTF-8", string, strlen(string), &res, &reslen)) {
+		perror("pdfgrep: Failed to remove accents: ");
+		return strdup(string);
 	}
 
-	return i;
+	return res;
 }
 
-int next_word_right(char *string, int index, int buflen)
+void simple_unac_free(char *string)
 {
-	int i = 0;
-	int in_whitespace;
-
-	if (index > buflen || string[index] == '\n')
-		return -1;
-
-	in_whitespace = isspace(string[index]);
-	while (index < buflen && string[index] != '\n') {
-		if (in_whitespace) {
-			if (!isspace(string[index]))
-				in_whitespace = 0;
-		} else {
-			if (isspace(string[index]))
-				break;
-		}
-		i++;
-		index++;
-	}
-
-	return i;
+	if (use_unac)
+		free(string);
 }
+#endif
 
-void print_context(char *string, int pos, int matchend, int buflen)
+int search_in_document(poppler::document *doc, const std::string &filename, regex_t *needle)
 {
-	int a = pos;
-	int b = matchend;
-	int chars_left = context;
-
-	int left = next_word_left(string, a-1);
-	int right = next_word_right(string, b, buflen);
-
-	while (true) {
-		if ((left < 0 || left > chars_left)
-			&& (right < 0 || right > chars_left))
-			break;
-
-		if (right < 0 || right > chars_left ||
-				(left > 0 && left <= chars_left
-				 && (pos - a) + left < (b - matchend) + right)) {
-			a -= left;
-			chars_left -= left;
-			left = next_word_left(string,a-1);
-		} else {
-			b += right;
-			chars_left -= right;
-			right = next_word_right(string, b, buflen);
-		}
-	}
-
-	putsn(string, a, pos);
-
-	with_color(highlight_color,
-		putsn(string, pos, matchend);
-	);
-
-	putsn(string, matchend, b);
-}
-
-void print_context_line(char *string, int pos, int matchend, int buflen)
-{
-	int a = pos;
-	int b = matchend;
-
-	while (a >= 0 && string[a] != '\n')
-		a--;
-	a++;
-	
-	while (b < buflen && string[b] != '\n')
-		b++;
-
-
-	putsn(string, a, pos);
-
-	with_color(highlight_color,
-		putsn(string, pos, matchend);
-	);
-
-	putsn(string, matchend, b);
-}
-
-int search_in_document(PDFDoc *doc, regex_t *needle)
-{
-	GooString *filename = doc->getFileName();
-	struct stream *stream = make_stream();
 	regmatch_t match[] = {{0, 0}};
 	int count_matches = 0;
-	int length = 0, oldcontext;
-	TextOutputDev *text_out = new TextOutputDev(write_to_stream, stream,
-			gFalse, gFalse);
+	int length = 0;
+	struct context cntxt = {context, 0, (char*)filename.c_str(), 0, &outconf};
 
-	if (!text_out->isOk()) {
-		if (!quiet) {
-			fprintf(stderr, "pdfgrep: Could not search %s\n", filename->getCString());
+
+
+	for (int i = 1; i <= doc->pages(); i++) {
+		std::auto_ptr<poppler::page> doc_page(doc->create_page(i - 1));
+		if (!doc_page.get()) {
+			if (!quiet) {
+				fprintf(stderr, "pdfgrep: Could not search in page %d of %s\n", i, filename.c_str());
+			}
+			continue;
 		}
-		goto clean;
-	}
 
+		cntxt.pagenum = i;
 
-	for (int i = 1; i <= doc->getNumPages(); i++) {
-		doc->displayPage(text_out, i,
-				72.0, 72.0, 0,
-				gTrue, gFalse, gFalse);
-
-		stream->buf[stream->charpos] = '\0';
+		poppler::byte_array str = doc_page->text().to_utf8();
+		str.resize(str.size() + 1, '\0');
+#ifdef HAVE_UNAC
+		char *unac_str = simple_unac(&str[0]);
+		char *str_start = unac_str;
+#else
+		char *str_start = &str[0];
+#endif
 		int index = 0;
-		
-		while (!regexec(needle, stream->buf+index, 1, match, 0)) {
+		struct match mt = {str_start, str.size()};
+
+		while (!regexec(needle, str_start+index, 1, match, 0)) {
+			regmatch_to_match(match[0], index, &mt);
+
 			count_matches++;
 			if (quiet) {
+#ifdef HAVE_UNAC
+				simple_unac_free(unac_str);
+#endif
 				goto clean;
 			} else if (!count) {
-				if (print_filename) {
-					with_color(filename_color,
-						printf("%s", filename->getCString()););
-					with_color(seperator_color, printf(":"););
-				}
-				if (print_pagenum) {
-					with_color(pagenum_color,
-						printf("%d", i););
-					with_color(seperator_color, printf(":"););
-				}
-				if (print_filename || print_pagenum)
-					printf(" ");
-
 				switch (context) {
 				/* print whole line */
 				case -1:
-					print_context_line(stream->buf, index + match[0].rm_so,
-							index + match[0].rm_eo, stream->charpos);
+					print_context_line(&cntxt, &mt);
 					break;
 				/* try to guess the terminal width */
 				case -2:
@@ -314,26 +191,20 @@ int search_in_document(PDFDoc *doc, regex_t *needle)
 					 * filename:linenumber: */
 					length = 0;
 
-					if (print_filename)
-						length += 1 + filename->getLength();
-					if (print_pagenum)
+					if (outconf.filename)
+						length += 1 + filename.size();
+					if (outconf.pagenum)
 						length += 1 + (int)log10(i);
-					if (print_pagenum || print_filename)
-						length += 1;
+
 					length += match[0].rm_eo - match[0].rm_so;
 
-					oldcontext = context;
-					context = line_width - length;
+					cntxt.before = line_width - length;
 
-					print_context(stream->buf, index + match[0].rm_so,
-							index + match[0].rm_eo, stream->charpos);
-
-					context = oldcontext;
+					print_context_chars(&cntxt, &mt);
 					break;
 				/* print a fixed context */
 				default:
-					print_context(stream->buf, index + match[0].rm_so,
-							index + match[0].rm_eo, stream->charpos);
+					print_context_chars(&cntxt, &mt);
 					break;
 				}
 
@@ -345,24 +216,17 @@ int search_in_document(PDFDoc *doc, regex_t *needle)
 			index += match[0].rm_so + 1;
 		}
 
-
-		reset_stream(stream);
+#ifdef HAVE_UNAC
+		simple_unac_free(unac_str);
+#endif
 	}
 
 	if (count && !quiet) {
-		if (print_filename) {
-			with_color(filename_color,
-				printf("%s", filename->getCString()););
-			with_color(seperator_color, printf(":"););
-		}
+		print_line_prefix(&outconf, filename.c_str(), -1);
 		printf("%d\n", count_matches);
 	}
-	
-clean:
-	free(stream->buf);
-	free(stream);
-	delete text_out;
 
+clean:
 	return count_matches;
 }
 
@@ -418,12 +282,12 @@ void read_colors_from_env(const char* env_var)
 		GLOBAL_VAR = strdup(cur_value); /* set to new color */ \
     }
 		/* now check for known settings and set global colors */
-		PARSE_COLOR("mt", highlight_color)
-		else PARSE_COLOR("ms", highlight_color)
-		else PARSE_COLOR("mc", highlight_color)
-		else PARSE_COLOR("fn", filename_color)
-		else PARSE_COLOR("ln", pagenum_color)
-		else PARSE_COLOR("se", seperator_color)
+		PARSE_COLOR("mt", outconf.colors.highlight)
+		else PARSE_COLOR("ms", outconf.colors.highlight)
+		else PARSE_COLOR("mc", outconf.colors.highlight)
+		else PARSE_COLOR("fn", outconf.colors.filename)
+		else PARSE_COLOR("ln", outconf.colors.pagenum)
+		else PARSE_COLOR("se", outconf.colors.separator)
 #undef PARSE_COLOR
 	}
 	/* free our copy of the environment var */
@@ -432,17 +296,18 @@ void read_colors_from_env(const char* env_var)
 
 void set_default_colors()
 {
-	filename_color = strdup("35");
-	pagenum_color = strdup("32");
-	highlight_color = strdup("01;31");
-	seperator_color = strdup("34");
+	outconf.colors.filename = strdup("35");
+	outconf.colors.pagenum = strdup("32");
+	outconf.colors.highlight = strdup("01;31");
+	outconf.colors.separator = strdup("36");
 }
 
 void free_colors()
 {
-	free(filename_color);
-	free(pagenum_color);
-	free(highlight_color);
+	free(outconf.colors.filename);
+	free(outconf.colors.pagenum);
+	free(outconf.colors.highlight);
+	free(outconf.colors.separator);
 }
 
 void init_colors()
@@ -493,11 +358,11 @@ void print_help(char *self)
 " -c, --count\t\t\tPrint only a count of matches per file\n"
 " -C, --context NUM\t\tPrint at most NUM chars of context\n"
 "     --color WHEN\t\tUse colors for highlighting;\n"
-" -q, --quiet\t\t\tSuppress normal output\n"
 "\t\t\t\tWHEN can be `always', `never' or `auto'\n"
+" -q, --quiet\t\t\tSuppress normal output\n"
+" -R, --recursive\t\tSearch directories recursively\n"
 "     --help\t\t\tPrint this help\n"
-" -V, --version\t\t\tShow version information\n"
-, self);
+" -V, --version\t\t\tShow version information\n");
 }
 
 void print_version()
@@ -505,13 +370,82 @@ void print_version()
 	printf("This is %s version %s\n", PACKAGE, VERSION);
 }
 
+int is_dir(const std::string filename)
+{
+	struct stat st;
+
+	if (stat(filename.c_str(), &st) == 0 && S_ISDIR(st.st_mode))
+		return 1;
+	else
+		return 0;
+}
+
+int do_search_in_document(const std::string path, const std::string filename,
+			  regex_t *ptrRegex, bool check_excludes = true)
+{
+	if (check_excludes &&
+	    (!is_excluded(includes, filename) || is_excluded(excludes, filename)))
+		return 0;
+
+	std::auto_ptr<poppler::document> doc(poppler::document::load_from_file(path));
+
+	if (!doc.get() || doc->is_locked()) {
+		fprintf(stderr, "pdfgrep: Could not open %s\n",
+			path.c_str());
+		return 1;
+	}
+
+	if (search_in_document(doc.get(), path, ptrRegex) && quiet) {
+		exit(0);
+	}
+
+	return 0;
+}
+
+int do_search_in_directory(const std::string filename, regex_t *ptrRegex)
+{
+	DIR *ptrDir = NULL;
+	struct dirent *ptrDirent = NULL;
+
+	ptrDir = opendir(filename.c_str());
+	if (!ptrDir) {
+		fprintf(stderr, "pdfgrep: %s: %s\n", filename.c_str(),
+			strerror(errno));
+		return 1;
+	}
+
+	while(1) {
+		std::string path(filename);
+		errno = 0;
+		ptrDirent = readdir(ptrDir);    //not sorted, in order as `ls -f`
+		if (!ptrDirent)
+			break;
+
+		if (!strcmp(ptrDirent->d_name, ".") || !strcmp(ptrDirent->d_name, ".."))
+			continue;
+
+		path += "/";
+		path += ptrDirent->d_name;
+
+		if (is_dir(path)) {
+			do_search_in_directory(path, ptrRegex);
+		} else {
+			do_search_in_document(path, ptrDirent->d_name, ptrRegex);
+		}
+	}
+
+	closedir(ptrDir);
+
+	return 0;
+}
+
 int main(int argc, char** argv)
 {
 	regex_t regex;
 	init_colors();
 
 	while (1) {
-		int c = getopt_long(argc, argv, "icC:nhHVq",
+		int c = getopt_long(argc, argv, "icC:nrRhHVq",
 				long_options, NULL);
 
 		if (c == -1)
@@ -525,13 +459,17 @@ int main(int argc, char** argv)
 				print_version();
 				exit(0);
 			case 'n':
-				print_pagenum = 1;
+				outconf.pagenum = 1;
+				break;
+			case 'r':
+			case 'R':
+				f_recursive_search = 1;
 				break;
 			case 'h':
-				print_filename = 0;
+				outconf.filename = 0;
 				break;
 			case 'H':
-				print_filename = 1;
+				outconf.filename = 1;
 				break;
 			case 'i':
 				ignore_case = REG_ICASE;
@@ -543,11 +481,11 @@ int main(int argc, char** argv)
 				if (!optarg)
 					break;
 				if (!strcmp("always", optarg)) {
-					color = 2;
+					outconf.color = 2;
 				} else if (!strcmp("never", optarg)) {
-					color = 0;
+					outconf.color = 0;
 				} else {
-					color = 1;
+					outconf.color = 1;
 				}
 				break;
 			case 'C':
@@ -559,9 +497,20 @@ int main(int argc, char** argv)
 					}
 				}
 				break;
+			case EXCLUDE_OPTION:
+				exclude_add(excludes, optarg);
+				break;
+			case INCLUDE_OPTION:
+				exclude_add(includes, optarg);
+				break;
 			case 'q':
 				quiet = 1;
 				break;
+#ifdef HAVE_UNAC
+			case UNAC_OPTION:
+				use_unac = 1;
+				break;
+#endif
 			case '?':
 				/* TODO: do something here */
 				break;
@@ -575,11 +524,12 @@ int main(int argc, char** argv)
 		exit(2);
 	}
 
-	globalParams = new GlobalParams();
-	globalParams->setTextPageBreaks(gFalse);
-	globalParams->setErrQuiet(gTrue);
+	char *pattern = argv[optind++];
+#ifdef HAVE_UNAC
+	pattern = simple_unac(pattern);
+#endif
 
-	int error = regcomp(&regex, argv[optind++], REG_EXTENDED | ignore_case);
+	int error = regcomp(&regex, pattern, REG_EXTENDED | ignore_case);
 	if (error) {
 		char err_msg[256];
 		regerror(error, &regex, err_msg, 256);
@@ -587,34 +537,39 @@ int main(int argc, char** argv)
 		exit(2);
 	}
 
-	if (color == 1 && !isatty(STDOUT_FILENO))
-		color = 0;
-	if (color)
-		read_colors_from_env("GREP_COLORS");
+	bool color_tty = isatty(STDOUT_FILENO) && getenv("TERM") &&
+		strcmp(getenv("TERM"), "dumb");
+	if (outconf.color == 1 && !color_tty) {
+		outconf.color = 0;
+	}
 
-	if (print_filename < 0)
-		print_filename = (argc - optind) == 1 ? 0 : 1;
+	if (outconf.color) read_colors_from_env("GREP_COLORS");
+
+	if (outconf.filename < 0) {
+		if ((argc - optind) == 1 && !is_dir(argv[optind])) {
+			outconf.filename = 0;
+		} else
+			outconf.filename = 1;
+	}
 
 	if (isatty(STDOUT_FILENO))
 		line_width = get_line_width();
 
+	if (excludes_empty(includes))
+		exclude_add(includes, "*.pdf");
+
 	error = 0;
 
 	for (int i = optind; i < argc; i++) {
-		GooString *s = new GooString(argv[i]);
-		PDFDoc *doc = new PDFDoc(s);
-		
-		if (!doc->isOk()) {
-			fprintf(stderr, "pdfgrep: Could not open %s\n", argv[i]);
+		const std::string filename(argv[i]);
+
+		if (!is_dir(filename)) {
+			do_search_in_document(filename, filename, &regex,false);
+		} else if (f_recursive_search) {
+			do_search_in_directory(filename, &regex);
+		} else { // TODO: report errors
 			error = 1;
-			continue;
 		}
-
-		if (search_in_document(doc, &regex) && quiet) {
-			exit(0);
-		}
-
-		delete doc;
 	}
 
 	if (error) {
